
# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

from __future__ import absolute_import
import json

version_json = '''
{
 "date": "2021-08-17T09:48:13+0200",
 "dirty": false,
 "error": null,
 "full-revisionid": "d4d4b0b2857d3fda4c177ac2066065f04a8877fc",
 "version": "2.2.1"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
